// Uses the Result enum

use std::error::Error;
use std::fs::{self, File};
use std::io::{self, ErrorKind, Read};

// Can actually return a result.
// Need final line to be Ok(())
fn main() -> Result<(), Box<dyn Error>> {
    let greeting_file_result = File::open("hello.txt");

    // let greeting_file = match greeting_file_result {
    //     Ok(file) => file,
    //     Err(error) => panic!("Problem opening file: {error:?}"),
    // };

    // Matching on different errors
    // let greeting_file = match greeting_file_result {
    //     Ok(file) => file,
    //     Err(error) => match error.kind() {
    //         std::io::ErrorKind::NotFound => match File::create("hello.txt") {
    //             Ok(fc) => fc,
    //             Err(e) => panic!("Problem creating file: {e:?}"),
    //         },
    //         other_error => {
    //             panic!("Problem opening file: {other_error:?}");
    //         }
    //     },
    // };

    // Using closures this time
    // let greeting_file = File::open("hello.txt").unwrap_or_else(|error| {
    //     if error.kind() == ErrorKind::NotFound {
    //         File::create("hello.txt").unwrap_or_else(|error| {
    //             panic!("Problem creating file: {error:?}");
    //         })
    //     } else {
    //         panic!("Problem opening file: {error:?}");
    //     }
    // });

    // shortcuts for panic on error
    // let greeting_file = File::open("hello.txt").unwrap();

    let greeting_file =
        File::open("hello.txt").expect("hello.txt should be included in this project");

    // propagating errors
    // instead of handling error within function itself, return
    // error to the calling code so it can decide what to do
    read_username_from_file();

    // A shortcut for propapating errors: the ? Operator
    read_username_from_file_question();

    // The code below errors
    // The ? can only be used in a function that returns Option or Result.
    // Similar to raising into a Monad in haskell with the pure function in a do statement.
    // let greeting_file = File::open("hello.txt")?;

    println!("Should be Some('a'): {:?}", last_char_of_first_line("line"));
    println!("Should be None: {:?}", last_char_of_first_line(""));

    Ok(())
}

fn read_username_from_file() -> Result<String, io::Error> {
    let username_file_result = File::open("hello.txt");

    let mut username_file = match username_file_result {
        Ok(file) => file,
        Err(e) => return Err(e),
    };

    let mut username = String::new();

    match username_file.read_to_string(&mut username) {
        Ok(_) => Ok(username),
        Err(e) => Err(e),
    }
}

fn read_username_from_file_question() -> Result<String, io::Error> {
    // let mut username_file = File::open("hello.txt")?;
    // let mut username = String::new();
    // username_file.read_to_string(&mut username)?;
    // Ok(username)

    // Can chain function calls
    // let mut username = String::new();
    // File::open("hello.txt")?.read_to_string(&mut username)?;
    // Ok(username)

    // Even faster way to write this
    // Just reads a file into a string
    fs::read_to_string("hello.txt")
}

fn last_char_of_first_line(text: &str) -> Option<char> {
    text.lines().next()?.chars().last()
}
